--_________________________________________________________________________________________________
--                                                                                                |
--                                       |masterFIP core|                                         |
--                                                                                                |
--                                         CERN,BE/CO-HT                                          |
--________________________________________________________________________________________________|

---------------------------------------------------------------------------------------------------
--                                                                                                |
--                                         masterfip_rx                                           |
--                                                                                                |
---------------------------------------------------------------------------------------------------
-- File         masterfip_rx.vhd                                                                  |
--                                                                                                |
-- Description  The unit groups the main actions that regard FIELDRIVE data reception.            |
--              Figure 1 shows the main units/processes; the units RX DESERIALIZER, RX OSC and    |
--              RX DEGLITCHER come unmodified from the nanoFIP project.                           |
--              Figure 2 shows the WorldFIP frame structure; note that the fmc_masterfip_core     |
--              ignores completely the notion of PDU_TYPE, LGTH, MPS, etc fields inside the       |
--              PAYLOAD part of the frame and is not checking them at reception.                  |
--              It is the processor (MT) that is handling individually the bytes inside the       |
--              PAYLOAD frame field. On the other hand the FSS, CRC and FES fields are checked    |
--              and validated by the masterfip_rx.                                                |
--              As long as the rx_rst_i is not activated the deserializer probes the bus          |
--              looking for a FSS; after the FSS detection, the deserialized bytes are packed     |
--              one-by-one in 32-bit words to be provided to the MT. Upon the detection of a FES  |
--              either the rx_fss_crc_fes_ok_p_o or the rx_crc_wrong_p_o is activated to signal to|
--              MT for the end of a correct or erroneous frame.                                   |
--                                                                                                |
--              RX OSC              for the clock recovery                                        |
--                                                                                                |
--              RX                  for the filtering of the input FD_RXD                         |
--                                                                                                |
--              RX DESERIALIZER     for the bytes retrieval; also detects FSS/FES & checks CRC    |
--                                                                                                |
--              BYTES_C             for the counting of the retrieved bytes                       |
--                                                                                                |
--              CREATE 32bit WORDS  for the formation of 32-bit words to be provided to the       |
--                                  processor (Mock Turtle for example)                           |
--                                                                                                |
--                                           Mock Turtle                                          |
--                     ___________________________________________________________                |
--                    |                                                           |               |
--                    |   _________       _______________________                 |               |
--                    |  |         |     |                       |                |               |
--                    |  | BYTES_C |     |   CREATE 32bit WORDS  |                |               |
--                    |  |_________|     |_______________________|                |               |
--                    |      /\                     /\                 _________  |               |
--                    |   _______________________________________     |         | |               |
--                    |  |                                       |    |         | |               |
--                    |  |             RX DESERIALIZER           |    |  RX OSC | |               |
--                    |  |                                       |  < |         | |               |
--                    |  |_______________________________________|    |         | |               |
--                    |                     /\                        |_________| |               |
--                    |   _______________________________________                 |               |
--                    |  |                                       |                |               |
--                    |  |              RX DEGLITCHER            |                |               |
--                    |  |_______________________________________|                |               |
--                    |                                                           |               |
--                    |___________________________________________________________|               |
--                                                 /\                                             |
--              ___________________________________________________________________               |
--              0_____________________________FIELDBUS______________________________0             |
--                                                                                                |
--           ___________ ______  ________________________________________  ___________ _______    |
--          |____FSS____|_CTRL_||_____________..DATA/PAYLOAD..___________||____CRC____|__FES__|   |
--                                                                                                |
--                                  Figure 2: WorldFIP Frame structure                            |
--                                                                                                |
-- Authors      Evangelia Gousiou     (Evangelia.Gousiou@cern.ch)                                 |
---------------------------------------------------------------------------------------------------

---------------------------------------------------------------------------------------------------
--                               GNU LESSER GENERAL PUBLIC LICENSE                                |
--                              ------------------------------------                              |
-- This source file is free software; you can redistribute it and/or modify it under the terms of |
-- the GNU Lesser General Public License as published by the Free Software Foundation; either     |
-- version 2.1 of the License, or (at your option) any later version.                             |
-- This source is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;       |
-- without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.      |
-- See the GNU Lesser General Public License for more details.                                    |
-- You should have received a copy of the GNU Lesser General Public License along with this       |
-- source; if not, download it from http://www.gnu.org/licenses/lgpl-2.1.html                     |
---------------------------------------------------------------------------------------------------



--=================================================================================================
--                                       Libraries & Packages
--=================================================================================================

-- Standard library
library IEEE;
use IEEE.STD_LOGIC_1164.all; -- std_logic definitions
use IEEE.NUMERIC_STD.all;    -- conversion functions
use IEEE.math_real.all;
-- Specific library
library work;
use work.masterFIP_pkg.all;  -- definitions of types, constants, entities
use work.wf_package.all;


--=================================================================================================
--                              Entity declaration for masterfip_rx
--=================================================================================================
entity masterfip_rx is port(
  -- INPUTS
    clk_i                 : in std_logic;   -- only one clk domain

    rst_i                 : in std_logic;   -- core rst, synched with clk_i

    rx_rst_i              : in std_logic;   -- dedicated rx reset during transmission OR
                                            -- reset pulse from the processor when for eg. a
                                            -- frame is rejected during reception
                                            -- (ex: RP_DAT > 133 bytes, wrong CTRL byte)

    speed_i               : in std_logic_vector(1 downto 0); -- WorldFIP bit rate

    rx_d_a_i              : in std_logic;   -- FielDrive receiver data


  -- OUTPUTS
    rx_byte_ready_p_o     : out std_logic;  -- pulse indicating a new retrieved byte

    rx_byte_o             : out std_logic_vector(C_BYTE_WIDTH-1 downto 0);
                                            -- currently retrieved byte

    rx_byte_index_o       : out std_logic_vector(C_FRAME_BYTES_CNT_LGTH-1 downto 0);
                                            -- index of currently retrieved byte
                                            -- counting starts after FSS; it includes
                                            -- the CTRL, DATA, CRC and FES fields;
                                            -- counting starts from 0 and normally
                                            -- the value should not exceed 265

    rx_word_index_o       : out std_logic_vector(C_FRAME_WORDS_CNT_LGTH-1 downto 0);
                                            -- index of currently retrieved word
                                            -- counting starts from 0 and normally
                                            -- the value should not exceed 66

    rx_ctrl_byte_o        : out std_logic_vector(C_BYTE_WIDTH-1 downto 0); -- frame CTRL byte

    rx_frame_o            : out rx_frame_t; -- frame DATA bytes
                                            -- structure with 67 words of 32-bit each = 268 bytes
                                            -- able to house a frame of max length.
                                            -- Upon the rx_fss_crc_fes_ok_p_o the processor needs
                                            -- to read the rx_ctrl_byte and rx_byte_index_o - 4
                                            -- bytes from the rx_frame_o. The content of rx_frame_o
                                            -- changes upon the end of reception of a new frame
                                            -- (i.e. a new rx_fss_crc_fes_ok_p_o).

    rx_fss_crc_fes_ok_p_o : out std_logic;  -- indication of a frame with correct FSS, FES & CRC;
                                            -- pulse upon FES detection

    rx_crc_wrong_p_o      : out std_logic;  -- indication of a frame with wrong CRC; pulse upon FES

    rx_fss_received_p_o   : out std_logic); -- pulse upon FSS detection (ID/ RP_DAT)

end entity masterfip_rx;


--=================================================================================================
--                                    architecture declaration
--=================================================================================================
architecture struc of masterfip_rx is

  -- wf_rx_osc
  signal rx_osc_rst, adjac_bits_window   : std_logic;
  signal signif_edge_window              : std_logic;
  signal sample_bit_p                    : std_logic;
  signal sample_manch_bit_p              : std_logic;
  -- wf_rx_deglitcher
  signal fd_rxd_filt, rxd_filt_edge_p    : std_logic;
  signal fd_rxd_filt_f_edge_p            : std_logic;
  signal fd_rxd_filt_r_edge_p            : std_logic;
  -- wf_rx_deserializer
  signal rx_byte_ready_p                 : std_logic;
  signal rx_fss_crc_fes_ok_p             : std_logic;
  signal rx_byte                         : std_logic_vector  (C_BYTE_WIDTH-1 downto 0);
  -- retrieved bytes into 32-bit regs
  signal byte0, byte1, byte2, byte3      : std_logic_vector(C_BYTE_WIDTH-1 downto 0);
  signal zero                            : std_logic_vector(C_BYTE_WIDTH-1 downto 0) := (others => '0');
  signal word32_num                      : integer range 0 to C_MAX_FRAME_WORDS-1;
  -- bytes counter
  signal rx_byte_index, rx_byte_index_d1 : std_logic_vector(C_FRAME_BYTES_CNT_LGTH-1 downto 0);
  signal bytes_c_rst                     : std_logic;


--=================================================================================================
--                                       architecture begin
--=================================================================================================
begin

---------------------------------------------------------------------------------------------------
--                            Deserializer as in the nanoFIP project                             --
---------------------------------------------------------------------------------------------------

  cmp_rx_deglitcher: wf_rx_deglitcher
  port map(
    uclk_i                 => clk_i,
    nfip_rst_i             => rx_rst_i,
    fd_rxd_a_i             => rx_d_a_i,
  -----------------------------------------------------------------
    fd_rxd_filt_o          => fd_rxd_filt,
    fd_rxd_filt_edge_p_o   => rxd_filt_edge_p,
    fd_rxd_filt_f_edge_p_o => fd_rxd_filt_f_edge_p);
  -----------------------------------------------------------------
    fd_rxd_filt_r_edge_p <= rxd_filt_edge_p and (not fd_rxd_filt_f_edge_p);

--  --  --  --  --  --  --  --  --  --  --  --  --  --  --  --  --  --  --  --  --  --  --  --  --
  cmp_rx_deserializer: wf_rx_deserializer
  port map(
    uclk_i                  => clk_i,
    nfip_rst_i              => rst_i,
    rx_rst_i                => rx_rst_i,
    fd_rxd_f_edge_p_i       => fd_rxd_filt_f_edge_p,
    fd_rxd_r_edge_p_i       => fd_rxd_filt_r_edge_p,
    fd_rxd_i                => fd_rxd_filt,
    sample_bit_p_i          => sample_bit_p,
    sample_manch_bit_p_i    => sample_manch_bit_p,
    signif_edge_window_i    => signif_edge_window,
    adjac_bits_window_i     => adjac_bits_window,
  -----------------------------------------------------------------
    byte_o                  => rx_byte,
    byte_ready_p_o          => rx_byte_ready_p,
    fss_crc_fes_ok_p_o      => rx_fss_crc_fes_ok_p,
    rx_osc_rst_o            => rx_osc_rst,
    fss_received_p_o        => rx_fss_received_p_o,
    crc_wrong_p_o           => rx_crc_wrong_p_o);
  -----------------------------------------------------------------
  rx_byte_ready_p_o         <= rx_byte_ready_p;
  rx_byte_o                 <= rx_byte;
  rx_fss_crc_fes_ok_p_o     <= rx_fss_crc_fes_ok_p;

--  --  --  --  --  --  --  --  --  --  --  --  --  --  --  --  --  --  --  --  --  --  --  --  --
  cmp_rx_osc: wf_rx_osc
  port map(
    uclk_i                  => clk_i,
    rate_i                  => speed_i,  -- or rx_rst??
    nfip_rst_i              => rst_i,
    fd_rxd_edge_p_i         => rxd_filt_edge_p,
    rx_osc_rst_i            => rx_osc_rst,
  -----------------------------------------------------------------
    rx_manch_clk_p_o        => sample_manch_bit_p,
    rx_bit_clk_p_o          => sample_bit_p,
    rx_signif_edge_window_o => signif_edge_window,
    rx_adjac_bits_window_o  => adjac_bits_window);
  -----------------------------------------------------------------


---------------------------------------------------------------------------------------------------
--                                         bytes counter                                         --
---------------------------------------------------------------------------------------------------
  cmp_rx_bytes_cnt: incr_counter
  generic map(g_counter_lgth => C_FRAME_BYTES_CNT_LGTH)
  port map(
    clk_i             => clk_i,
    counter_reinit_i  => bytes_c_rst,
    counter_incr_i    => rx_byte_ready_p,
    counter_is_full_o => open,
    -------------------------------------------------------
    counter_o         => rx_byte_index);
    -------------------------------------------------------
  bytes_c_rst         <= '1' when (rst_i = '1' or rx_rst_i = '1') else '0';
  rx_byte_index_o     <= rx_byte_index;


---------------------------------------------------------------------------------------------------
--                      combination of four retrieved bytes to a 32-bit word                     --
---------------------------------------------------------------------------------------------------
  create_32bit_words: process (clk_i)
  begin
    if rising_edge (clk_i) then
      if rst_i = '1' or rx_rst_i = '1' then
        byte0   <= (others => '0');
        byte1   <= (others => '0');
        byte2   <= (others => '0');
        byte3   <= (others => '0');
        rx_ctrl_byte_o <= (others  => '0');
        else
        if rx_byte_ready_p = '1' then
          if unsigned(rx_byte_index) = resize(unsigned(c_CTRL_BYTE_INDEX),C_FRAME_BYTES_CNT_LGTH) then
            rx_ctrl_byte_o <= rx_byte;
          else
            byte0          <= rx_byte;
            byte1          <= byte0;
            byte2          <= byte1;
            byte3          <= byte2;
          end if;
        end if;
      end if;
    end if;
   end process;


---------------------------------------------------------------------------------------------------
--                        Storage of consumed bytes to the cons_frame regs                       --
---------------------------------------------------------------------------------------------------
-- transfer 32bit words to the cons_frame registers

  delay: process (clk_i)
  begin
    if rising_edge (clk_i) then
      if rst_i = '1' or rx_rst_i = '1' then
        rx_byte_index_d1   <= (others =>'0'); -- needed for synching
      else
        if rx_byte_ready_p = '1' then
          rx_byte_index_d1 <= rx_byte_index;
        end if;
      end if;
    end if;
  end process;

--  --  --  --  --  --  --  --  --  --  --  --  --  --  --  --  --  --  --  --  --  --  --  --  --
  data_transfer_to_regs: process (clk_i)
  begin
    if rising_edge (clk_i) then
      if rst_i = '1' or rx_rst_i = '1' then
        word32_num                 <= 0;
        rx_frame_o                 <= (others =>(others =>'0'));

      else
        -- upon rx_fss_crc_fes_ok_p, the last 32bit word will contain for sure CRC0 and CRC1
        -- and it may also contain one or two bytes of data.
        -- the last word will always include as last bytes, byte1 and byte0: the two CRC bytes.
        -- it could also include one or two useful data bytes
        if rx_fss_crc_fes_ok_p = '1' then
          if word32_num = 0 then                -- only in the case of RP_FIN, where there are not enough
                                                -- bytes to create a word; needed for keeping the MT sw generic
            word32_num             <= word32_num + 1;
            rx_frame_o(word32_num) <= byte0 & byte1 & byte2 & byte3;

          elsif (unsigned(rx_byte_index)-2) mod 4 = 3 then -- [CRC|CRC|BYTE|BYTE]
            word32_num             <= word32_num + 1;
            rx_frame_o(word32_num) <= byte0 & byte1 & byte2 & byte3; -- byte 3 and byte 2 are useful

          elsif (unsigned(rx_byte_index)-2) mod 4 = 2 then -- [0|CRC|CRC|BYTE]
            word32_num             <= word32_num + 1;
            rx_frame_o(word32_num) <= zero & byte0 & byte1 & byte2; -- one useful data byte: byte2 last byte
                                                 
          end if;                                          -- Note: for [CRC|BYTE|BYTE|BYTE], upon
                                                           -- rx_fss_crc_fes_ok_p a new word has been created


        elsif (rx_byte_ready_p = '1' and unsigned(rx_byte_index_d1) > 0 and unsigned(rx_byte_index_d1) mod 4 = 0) then
          word32_num             <= word32_num + 1;
          rx_frame_o(word32_num) <= byte0 & byte1 & byte2 & byte3;
        end if;
      end if;
    end if;
  end process;


  rx_word_index_o <= std_logic_vector(to_unsigned(word32_num,C_FRAME_WORDS_CNT_LGTH));


end architecture struc;
--=================================================================================================
--                                        architecture end
--=================================================================================================
---------------------------------------------------------------------------------------------------
--                                      E N D   O F   F I L E
---------------------------------------------------------------------------------------------------